(* Name     : Domain.ML - theorems for Domain.thy
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 20-Oct-94
   updated  : 28-Oct-94 proofs for direct product
   updated  : 03-Nov-94 proofs for power set
   updated  : 04-Jan-95 adapted to new version of automatic proofs
			sharpness by theorem of Desharnais
*)

open Domain;


val  g =  g_ Domain.thy;
val pg = pg_ Domain.thy;


val Domain1_ss = Functionals_ss addsimps [
				product1_def, product2_def, 
				quasiprod_def, product3_def,
				product_def, prod_pi_rho,
				powerset1_def, powerset2_def,
				powerset_def, powerset_eps];

val dpre_tac = pre_tac Domain1_ss AtomRA_ss;
fun dpg str = pg str [dpre_tac 1, REPEAT(rfast_tac 1)];


(**** direct product *********************************************************)

val pr_pr1 = pg "Product(p,r) ==> Product1(p,r)" [
				rewrite_goals_tac [product_def,quasiprod_def],
				fast_tac HOL_cs 1];
val pr_univalent1 = pg "Product(p,r) ==> Univalent(p)" [
				dtac pr_pr1 1,
				rewtac product1_def,
				fast_tac HOL_cs 1];
val pr_univalent2 = pg "Product(p,r) ==> Univalent(r)" [
				dtac pr_pr1 1,
				rewtac product1_def,
				fast_tac HOL_cs 1];
val pr_pr2 = pg "Product(p,r) ==> Product2(p,r)" [
				rewrite_goals_tac [product_def,quasiprod_def],
				fast_tac HOL_cs 1];
val pr_pr3 = pg "Product(p,r) ==> Product3(p,r)" [
				rewtac product_def,
				fast_tac HOL_cs 1];
				
val pr2_pairs = dpg "Product2(p,r) ==> p^r = L & r^p = L";
val pr_pairs = pr_pr2 RS pr2_pairs;

val pr2_surjective1 = dpg "Product2(p,r) ==> Surjective(p)";
val pr_surjective1 = pr_pr2 RS pr2_surjective1;
val product_2 = pg "Product(p,r) = Product(r,p)" [
				simp_tac Domain1_ss 1,
				safe_tac HOL_cs,
				REPEAT(EVERY[
					rtac conv_isoton1 1,
					asm_full_simp_tac Relation_ss 1])];
					

val pr13_function = pg "[| Univalent p; Product3(p,r) |] ==> Function(p)" [
				rewrite_goals_tac [function_def, product3_def],
				sstac total_2 1,
				safe_tac HOL_cs,
				mon2_tac meet_lb1 1,
				rtac incl_intro1 1,
				etac sym 1];
val pr_surjection1 = pg "Product(p,r) ==> Surjective(p) & Function(p)" [
				rtac conjI 1,
				etac pr_surjective1 1,
				rtac pr13_function 1,
				etac pr_univalent1 1,
				etac pr_pr3 1];
val pr_surjection2 = pg "Product(p,r) ==> Surjective(r) & Function(r)" [
				dtac (product_2 RS iffD1) 1,
				etac pr_surjection1 1];

val pr_lemma1 = pg "Product(p,r) ==> RpL = RL" [
				dtac pr_surjection1 1,
				rewrite_goals_tac [function_def, total_def],
				s_tac comp_assoc 1,
				safe_tac HOL_cs,
				etac ssubst 1,
				rtac refl 1];

val pr_univalent_lemma = pg 
"[| Product(p,r); Univalent(R1); Univalent(R2) |] ==> Univalent(R1p^R2r^)" [
				rewtac univalent_def,
				simp_tac Relation_ss 1,
				mon_tac comp_hdistr_meet1 1,
				mon_tac meet_isoton1 1,
				rtac comp_hdistr_meet2 1,
				mon_tac meet_isoton2 1,
				rtac comp_hdistr_meet2 1,
				mon_tac meet_isoton1 1,
				rtac meet_lb1 1,
				mon_tac meet_isoton2 1,
				rtac meet_lb2 1,
				mon_tac meet_isoton1 1,
				s_tac comp_assoc 1,
				mon_tac comp_isoton2 1,
				sstac comp_assoc 1,
				mon_tac comp_isoton1 1,
				atac 1,
				rtac incl_refl 1,
				rtac incl_refl 1,
				mon_tac meet_isoton2 1,
				s_tac comp_assoc 1,
				mon_tac comp_isoton2 1,
				sstac comp_assoc 1,
				mon_tac comp_isoton1 1,
				atac 1,
				rtac incl_refl 1,
				rtac incl_refl 1,
				asm_full_simp_tac Domain1_ss 1];
val pr_total_lemma = 
	 pg "[| Product(p,r); Total(R1); Total(R2) |] ==> Total(R1p^R2r^)" [
				rewtac total_def,
				forward_tac [pr_lemma1] 1,
				etac subst 1, (*4*back();*)
				s_tac meet_commut 1,
				sstac univalent_4a 1,
				asm_full_simp_tac Domain1_ss 1,
				s_tac comp_assoc 1,
				dtac pr_pairs 1,
				asm_full_simp_tac HOL_ss 1,
				rotate_tac 1 1,
				etac ssubst 1,
				asm_full_simp_tac Lattice_ss 1];
val pr_function_lemma = 
pg "[| Product(p,r); Function(R1); Function(R2) |] ==> Function(R1p^R2r^)" [
				rewtac function_def,
				fast_tac (HOL_cs addIs [pr_univalent_lemma,
							pr_total_lemma]) 1];

val pr_univalent_sharp = pg "[| Product(p,r); Univalent(R1); Univalent(R2)|] \
			   \==> (R1p^R2r^)(pS1rS2) = R1S1R2S2" [
				rtac sym 1,
				sstac univalent_distr 1,
				eatac pr_univalent_lemma 1,
				atac 1,
				sstac comp_assoc 1,
				sstac comp_assoc 1,
				sstac univalent_4a 1,
				asm_full_simp_tac Domain1_ss 1,
				s_tac meet_commut 1, (*3*back();*)
				sstac univalent_4a 1,
				asm_full_simp_tac Domain1_ss 1,
				REPEAT(s_tac comp_assoc 1),
				dtac pr_pairs 1,
				asm_full_simp_tac HOL_ss 1,
				sstac predicate_1 1,
				sstac predicate_1 1,
				simp_tac Relation_ss 1,
				s_tac univalent_distr 1,
				atac 1,
				s_tac meet_assoc 1,
				s_tac univalent_distr 1,
				atac 1,
				simp_tac Relation_ss 1];

val pr_sharpness_lemma = 
	     pg "Product1(p,r) ==> ((Q1p^Q2r^)(pR1rR2)  Q1R1Q2R2)" [
				rewrite_goals_tac [product1_def,univalent_def],
				safe_tac HOL_cs,
				mon_tac comp_hdistr_meet1 1,
				rtac meet_isoton 1,
				m2tac "Q1p^" meet_lb1 1,
				s_tac comp_assoc 1,
				sstac comp_assoc 1, (* 1*back();*)
				matac 1,
				simp_tac Domain1_ss 1,
				m2tac "Q2r^" meet_lb2 1,
				s_tac comp_assoc 1,
				sstac comp_assoc 1, (* 1*back();*)
				matac 1,
				simp_tac Domain1_ss 1];


(**** sharpness by theorem of Desharnais *************************************)

(*
val desharnais_lemm = pg "[| Q1T^T  Q1; S^SQ2  Q2; T^TR1  R1 |] \
	\==> (S^Q1T^)(TR1(SQ2T(p^r))R2)  (Q1p^Q2r^)(pR1rR2)"[
				sstac comp_assoc 1,
				m1tac "SQ2Tp^r" dedekind_2b 1,
				s_tac comp_assoc 1,
				m1tac "(TR1(Tp^SQ2r^)(rR2))"
					dedekind_1b 1,
				simp_tac Relation_ss 1,
				mon_tac comp_isoton 1,
				rtac comp_hdistr_meet1 1,
				m2tac "Q1T^" meet_lb2 2,
				m2tac "S^" meet_lb1 2,
				mon_tac meet_isoton2 1,
				m2tac "pT^" meet_lb1 1,
				rtac incl_refl 1,
				rtac comp_isoton 1,
				simp_tac Relation_ss 1,
				matac 1,
				matac 1, (*1*back();*)
				rtac incl_refl 1,
				matac 1,
				simp_tac Relation_ss 1];
*)
val desharnais_lemma = pg "[| Q1^Q2R1R2^  p^r; Q1Q2R2R1^  S^T; \
			      \ Q1T^T  Q1; S^SQ2  Q2; T^TR1  R1 |] \
			   \==> Q1R1Q2R2  (Q1p^Q2r^)(pR1rR2)" [
				subst_tac "Q2R2" meet_idempot 1,
				s_tac meet_assoc 1,
				mtac dedekind_2a 1,
				subst_tac "Q1" meet_idempot 1,
				sstac meet_assoc 1,
				matac 1,
				mtac dedekind_2b 1,
				s_tac comp_assoc 1,
				mon_tac dedekind_1a 1,
				simp_tac Relation_ss 1,
				subst_tac "TR1" meet_idempot 1,
				sstac meet_assoc 1,
				m1tac "TR1(STQ1^)Q2R2" dedekind_2b 1,
				mtac comp_hdistr_meet2 1, (*1*back();*)
				sstac meet_assoc 1,
				s_tac comp_assoc 1, (*1*back();*)
				m1tac "T(Q1^Q2)TR1R2^" dedekind_1a 1,
				sstac comp_assoc 1,
				sstac comp_assoc 1,
				matac 1,
				matac 1,
(*
				eatac desharnais_lemm 1,
				atac 1];
*)
				sstac comp_assoc 1,
				m1tac "SQ2Tp^r" dedekind_2b 1,
				s_tac comp_assoc 1,
				m1tac "(TR1(Tp^SQ2r^)(rR2))"
					dedekind_1b 1,
				simp_tac Relation_ss 1,
				mon_tac comp_isoton 1,
				rtac comp_hdistr_meet1 1,
				m2tac "Q1T^" meet_lb2 2,
				m2tac "S^" meet_lb1 2,
				mon_tac meet_isoton2 1,
				m2tac "pT^" meet_lb1 1,
				rtac incl_refl 1,
				rtac comp_isoton 1,
				simp_tac Relation_ss 1,
				matac 1,
				matac 1, (*2*back();*)
				rtac incl_refl 1,
				matac 1,
				simp_tac Relation_ss 1];
val desharnais_lemma2 = pg "[| QuasiProd(p,r); Q1Q2R2R1^  PL;\
			   \ (P::('a,'b)rel)P^(Q2::('a,'b)rel)  Q2|]\
			   \==> Q1R1Q2R2  (Q1p^Q2r^)(pR1rR2)" [
				asm_full_simp_tac Domain1_ss 1,
				safe_tac HOL_cs,
				rtac desharnais_lemma 1,
				dtac (equal_L RS iffD2) 1,
				etac ssubst 1,
				rtac L_def 1,
				etac incl_trans 1,
				dtac conv_isoton2 1, (*3*back();*)
				asm_full_simp_tac Relation_ss 1,
				dtac (equal_L RS iffD2) 1, 
				etac subst 1,
				sstac comp_assoc 1, (*2*back();*)
				subst_tac "Pr^" conv_invol 1,
				rtac incl_refl 1,
				s_tac I_def2 1, (*1*back();*)
				s_tac comp_assoc 1,
				matac 1,
				rtac incl_refl 1,
				simp_tac Relation_ss 1,
				s_tac comp_assoc 1, (*2*back();*)
				matac 1,
				sstac I_def2 1,
				atac 1,
				matac 1,
				simp_tac Relation_ss 1];

val pr_homogen_sharp = 
pg "QuasiProd(p,r) ==> Q1R1(Q2::('a,'a)rel)R2  (Q1p^Q2r^)(pR1rR2)" [
				ratac desharnais_lemma2 1,
				sstac I_def1 1,
				rtac L_def 1,
				simp_tac Relation_ss 1];

val pr_difunct_sharp = pg "[| QuasiProd(p,r); Difunctional(Q2) |] \
			  \ ==> Q1R1Q2R2  (Q1p^Q2r^)(pR1rR2)" [
				ratac desharnais_lemma2 1,
				rewtac difunctional_def,
				atac 2,
				mon_tac meet_lb2 1,
				s_tac comp_assoc 1,
				rtac comp_isoton2 1,
				rtac L_def 1];

val pi_rho_difunct_sharp = 
	      pg "Difunctional(Q2) ==> (Q1ͷ^Q2͸^)(R1ԸR2)=Q1R1Q2R2" [
				cut_facts_tac [prod_pi_rho] 1,
				rtac incl_antisym 1,
				rtac pr_sharpness_lemma 1,
				rewtac product_def,
				safe_tac HOL_cs,
				eatac pr_difunct_sharp 2,
				rewtac quasiprod_def,
				fast_tac HOL_cs 1];


(**** power set **************************************************************)

val ps_function = pg "Powerset(e) ==> Function(SyQ(R^,e))" [
				rewrite_goals_tac [powerset_def,
						  powerset2_def, function_def],
				safe_tac HOL_cs,
				fast_tac HOL_cs 2,
				rewrite_goals_tac[powerset1_def,univalent_def],
				sstac conv_sym_quot 1,
				mon_tac sym_quot_red_lemma 1,
				atac 1];

val ps_eps_function = pg "Function(SyQ(R^,))" [
				rtac ps_function 1,
				rtac powerset_eps 1];

val ps_inversion = pg "Powerset(e) ==> SyQ(R^,e)e^ = R" [
				rewrite_goals_tac [powerset_def,powerset2_def],
				rtac (conv_equal RS iffD1) 1,
				simp_tac(Relation_ss addsimps[conv_sym_quot])1,
				fast_tac (HOL_cs addEs [sym_quot_pre2]) 1];

val ps_mp = pg "Powerset(e) ==> e(eR) = R" [
				rtac incl_antisym 1,
				rtac right_res_mp 1,
				rewrite_goals_tac [powerset_def,powerset2_def],
				subst_tac "R" sym_quot_pre2 1, (*2*back();*)
				s_tac conv_invol 1, (*1*back();*)
				fast_tac HOL_cs 1,
				simp_tac Domain1_ss 1,
				fast_tac Relation_cs 1];
val ps_mp_cmpl = pg "Powerset(e) ==> -e(-e-R) = -R" [
				rtac incl_antisym 1,
				rtac right_res_mp 1,
				rewrite_goals_tac [powerset_def,powerset2_def],
				subst_tac "-R" sym_quot_pre2 1, (*2*back();*)
				sstac sym_quot_cmpl 1,
				s_tac conv_invol 1, (*1*back();*)
				fast_tac HOL_cs 1,
				simp_tac Domain1_ss 1,
				s_tac conv_cmpl 1,
				fast_tac Relation_cs 1];

val Domain_ss = Domain1_ss addsimps [ps_inversion, ps_mp, ps_mp_cmpl];
