(* Name     : Lattice.ML - theorems for Lattice.thy
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 27-May-94
   updated  : 30-May-94 proofs for laws upto lower and upper bounds
   updated  : 06-Jun-94 proofs for laws upto distributivity
   updated  : 10-Jun-94 proofs for atom_in laws
   updated  : 13-Jun-94 proofs for complemet laws
   updated  : 20-Jun-94 proofs for inf/sup laws
   updated  : 06-Jul-94 proofs for inf/sup atomic laws
   updated  : 13-Jul-94 proofs for Complenetness laws, new operator sytax
   updated  : 28-Jul-94 new structure of classical sets: Atomize/Juggle
   updated  : 08-Aug-94 new structure of fast_tac: Simplification in step_tac
   updated  : 12-Aug-94 slow_tac2, new classical set: Infer-set AtomLI_cs
   updated  : 01-Dec-94 major simplifications of many proofs (Lattice_cs)
   updated  : 06-Dec-94 major simplifications of many proofs, rest
   updated  : 07-Feb-97 adaptions for new HOL version
*)

open Lattice;

val  g =  g_ Lattice.thy;
val pg = pg_ Lattice.thy;


(* this enables to have a look at the types of L and O, if show_types is set
by(asm_full_simp_tac(HOL_ss addsimps [cmpl_defL RS sym, cmpl_defO RS sym]) 1); 
*)


(**** for Simplifier *********************************************************)

val meet_absorb2 = pg "x(yx) = x" [ sstac join_commut 1,
				      rtac meet_absorb 1];
val join_absorb2 = pg "x(yx) = x" [ sstac meet_commut 1,
				      rtac join_absorb 1];

val meet_lcommut = pg "x(yz) = y(xz)" [
				s_tac meet_assoc 1,
				s_tac meet_commut 1, (*1*back();*)
				rtac (meet_assoc RS sym) 1];

val join_lcommut = pg "x(yz) = y(xz)" [
				s_tac join_assoc 1,
				s_tac join_commut 1, (*1*back();*)
				rtac (join_assoc RS sym) 1];

val meet_idempot = pg "xx  =  x" [
				s_tac meet_absorb 1, (*3*back();*)
				sstac join_absorb 1,
				rtac refl 1];

val join_idempot = pg "xx  =  x" [
				s_tac join_absorb 1, (*3*back();*)
				sstac meet_absorb 1,
				rtac refl 1];


(**** for Inclusion relation *************************************************)

val incl_equiv = pg "x  y = (xy = y)" [
				rewtac incl_def,
				rtac iffI 1,
				s_tac join_absorb 1, (*1*back();*)
				sstac join_commut 1,
				s_tac meet_commut 1,
				etac ssubst 1,
				rtac refl 1,
				s_tac  meet_absorb 1, (*1*back();*)
				etac ssubst 1,
				rtac refl 1];

val incl_refl  = pg "x  x" [rewtac incl_def, rtac meet_idempot 1];
val incl_trans = pg "[| x  y; y  z |] ==> x  z" [
				rewtac incl_def,
				etac subst 1,
				etac subst 1, (*1*back();*)
				sstac meet_assoc 1,
				rtac refl 1];
val incl_e_trans = pg "x  y & y  z ==> x  z" [
				rtac incl_trans 1,
				etac conjunct1 1,
				etac conjunct2 1];
val incl_antisym = pg "[| x  y; y  x |] ==> x = y" [
				rewtac incl_def,
				etac subst 1,
				etac subst 1,
				rtac meet_commut 1];

val incl_intro1 = pg "x=y ==> x  y" [fast_tac (HOL_cs addIs [incl_refl]) 1];
val incl_intro2 = pg "x=y ==> y  x" [fast_tac (HOL_cs addIs [incl_refl]) 1];
val incl_intro  = pg "x=y ==> x  y & y  x" [
				fast_tac (HOL_cs addIs [incl_refl]) 1];
val equal_equiv = pg "(x=y) = (x  y & y  x)" [
			fast_tac (HOL_cs addIs [incl_refl, incl_antisym]) 1];

fun mon_tac  thm = (rtac incl_trans) THEN' (rtac thm);
fun mon2_tac thm i = (rtac incl_trans i) THEN (rtac thm (i+1));

val set_ss = simpset_of "Set";
val Lattice1_ss = set_ss addsimps
  [meet_assoc,meet_commut,meet_lcommut, meet_absorb,meet_absorb2,meet_idempot,
   join_assoc,join_commut,join_lcommut, join_absorb,join_absorb2,join_idempot,
	   cmpl_defL, cmpl_defO, O_def, L_def, incl_refl];


(**** Lower and Upper Bounds *************************************************)

fun rev_thm thm rev str = pg str [s_tac rev 1, rtac thm 1];
fun rev_meet_thm thm = hd([thm] RL [meet_commut RS subst] );
fun rev_join_thm thm = hd([thm] RL [join_commut RS subst] );

val meet_lb1 = pg "xy  x" [simp_tac (Lattice1_ss addsimps [incl_equiv]) 1];
val meet_lb2 = rev_thm meet_lb1 meet_commut "xyy";

val join_ub1 = pg "x  xy" [simp_tac (Lattice1_ss addsimps [incl_def]) 1];
val join_ub2=  rev_thm join_ub1 join_commut "y  xy";

val meet_glb = pg "[| z  x; z  y |] ==> z  xy" [
				rewtac incl_def,
				s_tac meet_assoc 1,
				asm_full_simp_tac HOL_ss 1];
val meet_lbI1 = pg "x  z ==> xy  z" [etac (meet_lb1 RS incl_trans) 1];
val meet_lbI2 = pg "y  z ==> xy  z" [etac (meet_lb2 RS incl_trans) 1];
val meet_ubD = pg "z  xy ==> z  x & z  y" [
				rtac conjI 1,
				etac (meet_lb1 RSN (2,incl_trans)) 1,
				etac (meet_lb2 RSN (2,incl_trans)) 1];
val join_lub = pg "[| x  z; y  z |] ==> xy  z" [
				asm_full_simp_tac (Lattice1_ss addsimps
					[incl_equiv]) 1];
val join_ubI1 = pg "z  x ==> z  xy" [etac (join_ub1 RSN (2,incl_trans)) 1];
val join_ubI2 = pg "z  y ==> z  xy" [etac (join_ub2 RSN (2,incl_trans)) 1];
val join_lbD = pg "xy  z ==> x  z & y  z" [
				rtac conjI 1,
				etac (join_ub1 RS incl_trans) 1,
				etac (join_ub2 RS incl_trans) 1];

val Lattice1_cs = set_cs addSIs [incl_refl,
				 meet_glb, join_lub]
			 addIs [incl_antisym, meet_hdistr1,
				O_def, L_def,
				meet_lbI1, meet_lbI2, join_ubI1, join_ubI2,
				inf_lb RS bspec, inf_glb RS mp,
				sup_ub RS bspec, sup_lub RS mp]
			 addEs [incl_e_trans]
			 addDs [incl_intro, meet_ubD, join_lbD];



(**** for Distributivity, 1 **************************************************)

val meet_hdistr2 = pg "(xy)(xz)  x(yz)" [fast_tac Lattice1_cs 1];
val meet_distr = pg "x(yz) = (xy)(xz)" [fast_tac Lattice1_cs 1];
val join_distr = pg "x(yz) = (xy)(xz)" [
				simp_tac (Lattice1_ss addsimps [meet_distr]) 1,
				s_tac join_assoc 1,
				simp_tac Lattice1_ss 1];



(**** for Completeness, 1 ****************************************************)

val inf_lb1 = pg "aA ==> Inf A  a" [fast_tac Lattice1_cs 1];
val sup_ub1 = pg "aA ==> a  Sup A" [fast_tac Lattice1_cs 1];
val inf_antiton2 = pg "A <= B ==> Inf(B)  Inf A" [fast_tac Lattice1_cs 1];
val sup_isoton2  = pg "A <= B ==> Sup A  Sup(B)" [fast_tac Lattice1_cs 1];



(**** for Atomic *************************************************************)

val atom_in_is_atom = pg "ax ==> Atomic(a)" [
				rewtac atom_in_def,
				fast_tac HOL_cs 1];

val atom_in_refl = pg "Atomic(a) ==> aa" [
				rewtac atom_in_def,
				fast_tac Lattice1_cs 1];

val atom_in_refl_eq = pg "aa = Atomic a" [
				fast_tac (HOL_cs addIs 
				  [atom_in_is_atom,atom_in_refl]) 1];

val atom_not_in_cmpl = pg "a-a = False" [
			rewrite_goals_tac [atom_in_def, atom_def, incl_def],
				step_tac HOL_cs 1,
				asm_full_simp_tac Lattice1_ss 1];

val atom_in_atom = pg "[| ab ; Atomic(b) |] ==> a=b" [
				rewrite_goals_tac [atom_in_def, atom_def],
				fast_tac HOL_cs 1];

val atom_not_in_O = pg "aO = False" [
				rewrite_goals_tac [atom_in_def, atom_def],
			fast_tac(HOL_cs addDs [O_def RS incl_antisym]) 1];
val atom_in_L = pg "lL = Atomic(l)" [
				rewtac atom_in_def,
				fast_tac Lattice1_cs 1];



(**** for Atomicity **********************************************************)

val strong_atomic_lemma0 = pg "Lx=x"[fast_tac (Lattice1_cs addSIs [L_def]) 1];
val strong_atomic_lemma1 = pg "(x-y)y=y ==> x  y" [asm_full_simp_tac
	(Lattice1_ss addsimps [join_distr,strong_atomic_lemma0,incl_equiv]) 1];
val strong_atomic_lemma2 = pg "~(x  y) ==> x-y  O" [
				etac swap 1,
				dtac notnotD 1,
				rtac strong_atomic_lemma1 1,
				etac ssubst 1,
				fast_tac Lattice1_cs 1];
val strong_atomic_lemma3 = pg "[| Atomic(a) ; a  x; a  -x |] ==> False" [
				datac meet_glb 1,
				asm_full_simp_tac Lattice1_ss 1,
				rewtac atom_def,
			fast_tac(HOL_cs addDs [O_def RS incl_antisym]) 1];
val strong_atomic_lemma4 = pg "Sup({xa. xa  x})  x" [fast_tac Lattice1_cs 1];
val strong_atomic = pg "Sup({a. ax}) = x" [
				rewtac atom_in_def,
				rtac incl_antisym 1,
				rtac incl_trans 1,
				rtac strong_atomic_lemma4 2,
				rtac sup_isoton2 1,
				fast_tac set_cs 1,
	(* x  sup({a. ax}) *) rtac ccontr 1,
				dtac strong_atomic_lemma2 1,
				dtac (weak_atomic RS mp) 1,
				rewtac atom_in_def,
				step_tac HOL_cs 1,
				ratac strong_atomic_lemma3 1,
				fast_tac Lattice1_cs 2,
				fast_tac Lattice1_cs 1];

val atom_in_incl = pg "x  y = (a. ax --> ay)" [
				rtac iffI 1,
				s_tac strong_atomic 2,
				subst_tac "x" strong_atomic 2,
				rtac sup_isoton2 2,
				fast_tac set_cs 2,
				rewtac atom_in_def,
				fast_tac (HOL_cs addSEs [incl_trans]) 1];

val atom_in_meet = pg "mxy = (mx & my)" [
				rewtac atom_in_def,
				fast_tac Lattice1_cs 1];


val atom_split_join_lemma = pg "[|  x  . xO & xa = x --> x = a; aza |] \
		\ ==> azO"   [etac allE 1, fast_tac Lattice1_cs 1];
val unequal_trans = pg "[| x=y; zy |] ==> xz" [fast_tac HOL_cs 1];
val atom_split_join = pg "[| Atomic(a); a  xy |] ==> a  x | a  y" [
				etac swap2 1, (*1*back();*)
				rewrite_goals_tac [incl_def,atom_def],
				etac conjE 1,
				ra2tac unequal_trans 1,
				sstac meet_distr 1,
				rtac (O_def RSN (2,incl_antisym)) 1,
				rtac join_lub 1,
				etac atom_split_join_lemma 1,
				etac atom_split_join_lemma 2,
				fast_tac Lattice1_cs 1,
				fast_tac Lattice1_cs 1];
val atom_in_join = pg "jxy = (jx | jy)" [
				rewtac atom_in_def,
				step_tac Lattice1_cs 1,
				datac atom_split_join 1,
				fast_tac HOL_cs 1,
				fast_tac Lattice1_cs 1,
				fast_tac Lattice1_cs 1];

val atom_in_cmpl_lemma = pg "~(ax & a-x)" [
				s_tac atom_in_meet 1,
				simp_tac (HOL_ss addsimps 
					[cmpl_defO, atom_not_in_O]) 1];
val atom_in_cmpl = pg "c-x = (~cx & Atomic(c))" [
				rtac iffI 1,
				cut_facts_tac [atom_in_cmpl_lemma] 1,
				fast_tac (HOL_cs addDs [atom_in_is_atom]) 1,
				rewtac atom_in_def,
				safe_tac HOL_cs,
				dtac atom_split_join 1,
				sstac cmpl_defL 1,
				rtac L_def 1,
				fast_tac HOL_cs 1];


val AtomL_cs1 = set_cs	addSIs [atom_in_cmpl RS iffD2, atom_in_incl RS iffD2,
				atom_in_meet RS iffD2, atom_in_join RS iffD2]
			addIs  [atom_in_L RS iffD2, atom_in_refl, incl_antisym]
			addSEs [atom_in_is_atom]
			addSDs [atom_in_cmpl RS iffD1, atom_in_incl RS iffD1,
				atom_in_meet RS iffD1, atom_in_join RS iffD1]
			addDs  [atom_not_in_O RS iffD1,
				atom_not_in_cmpl RS iffD1, incl_intro];

val AtomLA1_ss = HOL_basic_ss addsimps [atom_in_cmpl, atom_in_incl,
					atom_in_meet, atom_in_join,
					atom_in_L,atom_not_in_O];

val AtomL1_css =(set_cs	addSEs [thin_True_rl],
		 HOL_basic_ss setSolver (K (K no_tac)) setmksimps 
		  (mksimps (filter (fn x =>fst x<>"op -->")mksimps_pairs))
		  (*addsimps [atom_in_is_atom]*)
		)
		addSEs2' [sym,atom_in_is_atom];

val atom_in_ball = pg "(z{f(x) |x. P(x)}. az) = (x. P(x) --> af(x))" [
				fast_tac set_cs 1];
val atom_in_bex = pg "(z{f(x) |x. P(x)}. az) = (x. P(x) & af(x))" [
				fast_tac set_cs 1];


local
val setcompr =pg"SetCompr f P = {f x |x. P x}"[rewtac setcompr_def,rtac refl 1];
in
fun pre_tac general_ss atomize_ss = EVERY' [
				asm_full_simp_tac general_ss,	
				TRY o rewr_tac (setcompr RS sym),
				IF_UNSOLVED o (full_simp_tac 
					(HOL_basic_ss addsimps [equal_equiv])),
				TRY o rewr_tac setcompr,
			 	IF_UNSOLVED o (full_simp_tac atomize_ss),
				TRY o SELECT_GOAL (safe_tac set_cs)];
end;

val l1pre_tac  = pre_tac Lattice1_ss AtomLA1_ss;
val l1step_tac = step_tac (op addss2 AtomL1_css);
val l1fast_tac = fast_tac (op addss2 AtomL1_css);
val l1slow_tac = slow_tac (op addss2 AtomL1_css);
fun l1pg str   = pg str [l1pre_tac 1, REPEAT (FIRSTGOAL l1fast_tac)];

val meet_O = l1pg "Ox = O";
val meet_L = l1pg "Lx = x";
val join_O = l1pg "Ox = x";
val join_L = l1pg "Lx = L";

val equal_O = l1pg "(x=O) = (xO)";
val equal_L = l1pg "(x=L) = (Lx)";

val meet_atoms = pg "Atomic(a) & Atomic(b) & a  b --> ab = O"
		[l1pre_tac 1, 
		 REPEAT(fast_tac (op addss2 AtomL1_css addDs [atom_in_atom])1)];


(**** for Supremum and Infimum ***********************************************)

val (inf_lb1_a, sup_ub1_a, inf_glb_a, sup_lub_a) =
let fun tacs thm  = [		rtac (atom_in_incl RS iffD1 RS allE) 1,
				fast_tac HOL_cs 2,
				etac thm 1]
in (
pg "[| aInf A; xA |] ==> ax" (tacs inf_lb1),
pg "[| ax; xA |] ==> aSup A" (tacs sup_ub1),
pg "[| al; xA. l  x |] ==> aInf A" (tacs (inf_glb RS mp)),
pg "[| aSup A; xA. x  u |] ==> au" (tacs (sup_lub RS mp))
) end;

val atom_in_inf = pg "aInf A = (Atomic(a) & (zA. az))" [
				step_tac AtomL_cs1 1,
				fast_tac (AtomL_cs1 addEs  [inf_lb1_a]) 1,
				rtac inf_glb_a 1,
				etac atom_in_refl 1,
				rewtac atom_in_def,
				fast_tac set_cs 1];

val meet_sup_lemma = l1pg "y  z-x = xy  z";
val meet_sup = pg "xSup A = Sup({xy | y. yA})" [
				rtac incl_antisym 1,
				s_tac meet_sup_lemma 1,
				rtac (sup_lub RS mp) 1,
				safe_tac set_cs,
				sstac meet_sup_lemma 1,
				fast_tac(Lattice1_cs addSIs[sup_ub RS bspec])1,
				fast_tac (AtomL_cs1 addEs 
					[sup_ub1_a, sup_lub_a]) 1];

val atom_in_sup_lemma0=pg "[| (~(xA. ax)); Atomic(a) |] ==> (xA. ~(ax))"[
				fast_tac AtomL_cs1 1];
val atom_in_sup_lemma1 = pg "[| ~(ax); Atomic(a) |] ==> axO"
		[l1pre_tac 1,
		 REPEAT(fast_tac (op addss2 AtomL1_css addDs [atom_in_atom])1)];
val atom_in_sup_lemma2 = pg "[| axO |] ==> ~(ax)"
		[l1pre_tac 1,
		 REPEAT(fast_tac (op addss2 AtomL1_css addIs [atom_in_refl])1)];
val atom_in_sup = pg "aSup A = ( zA. az)" [
				rtac iffI 1,
				rtac swap 1, rtac notI 1,etac notE 1,atac 1,
				rtac atom_in_sup_lemma2 1,
				sstac meet_sup 1,
				rtac (sup_lub RS mp) 1,
				dtac atom_in_sup_lemma0 1,
				etac atom_in_is_atom 1,
				fast_tac (set_cs addSEs
				      [atom_in_sup_lemma1, atom_in_is_atom]) 1,
				fast_tac (set_cs addIs [sup_ub1_a]) 1];

val atmomic_eq = pg "a z .(P(z) & az) --> P(a) ==> \
		\aSup({x.P(x)}) = (Atomic(a) & P(a))" [
				sstac atom_in_sup 1,
				fast_tac (set_cs addIs [atom_in_refl]
						 addEs [atom_in_is_atom]) 1];

val Lattice2_ss = Lattice1_ss addsimps [meet_O, meet_L, join_O, join_L,
					equal_O, equal_L];
val AtomL_cs2  = AtomL_cs1 addSIs [atom_in_inf RS iffD2, atom_in_sup RS iffD2]
			   addSDs [atom_in_inf RS iffD1, atom_in_sup RS iffD1];
val AtomLA2_ss = AtomLA1_ss addsimps [atom_in_inf , atom_in_sup,
				      atom_in_ball, atom_in_bex];
val AtomL2_css = AtomL1_css;

val l2pre_tac    = pre_tac Lattice2_ss AtomLA2_ss;
val l2step_tac   = step_tac (op addss2 AtomL2_css);
val l2fast_tac   = fast_tac (op addss2 AtomL2_css);
val l2slow_tac   = slow_tac (op addss2 AtomL2_css);
fun l2pg     str = pg str [l2pre_tac 1, REPEAT (FIRSTGOAL l2fast_tac)];
fun l2slowpg str = pg str [l2pre_tac 1, REPEAT (FIRSTGOAL l2slow_tac)];

(**** for Isoton *************************************************************)

val meet_isoton1 = l2pg "v  x ==> vy  xy";
val meet_isoton2 = l2pg "w  y ==> xw  xy";
val meet_isoton  = l2pg "[| v  x; w  y |] ==> (vw)  (xy)";

val join_isoton1 = l2pg "v  x ==> vy  xy";
val join_isoton2 = l2pg "w  y ==> xw  xy";
val join_isoton  = l2pg "[| v  x; w  y |] ==> (vw)  (xy)";


(**** for Distributivity, 2 **************************************************)

val join_inf = l2pg "xInf A = Inf({xy | y. yA})";


(**** for Completeness, 2 ****************************************************)

val sup_union = l2pg "Sup(A Un B) = Sup A  Sup(B)";
val inf_union = l2pg "Inf(A Un B) = Inf A  Inf(B)";

val inf_antiton = l2pg "(yA. xB. xy) ==> Inf(B)  Inf A";
val sup_isoton  = l2pg "(xA. yB. xy) ==> Sup A  Sup(B)";


val inf_empty = l2pg "Inf {} = L";
val inf_step  = l2pg "Inf(insert x A) = xInf A";
val inf_is_meet = l2pg "Inf {x,y} = xy";

val sup_empty =	l2pg "Sup {} = O";
val sup_step  = l2pg "Sup (insert x A) = xSup A";
val sup_is_join = l2pg "Sup {x,y} = xy";

val inf_incl_sup = l2pg 
		"L(O::'a::lattice) ==> ( x. xA) = Inf A  (Sup A::'a)";



(**** for Complement *********************************************************)

val cmpl_invol = l2pg "--x=x";

val cmpl_antiton = l2pg "xy = -y-x";
val cmpl_equal = l2pg "(-x=-y) = (x=y)";

val cmpl_meet = l2pg "-(xy)= -x-y";
val cmpl_join = l2pg "-(xy)= -x-y";

val cmpl_O = l2pg "-O=L";
val cmpl_L = l2pg "-L=O";

val incl_equivL = l2pg "xy = L  -xy";
val incl_equivO = l2pg "xy = x-y  O";

val cmpl_inf = l2pg "-Inf A = Sup({-x |x. xA})";
val cmpl_sup = l2pg "-Sup A = Inf({-x |x. xA})";


(*****************************************************************************)

val Lattice_ss = Lattice2_ss addsimps [cmpl_O, cmpl_L, cmpl_invol, cmpl_equal,
				     cmpl_meet, cmpl_join, cmpl_inf, cmpl_sup,
				     inf_empty, inf_step, sup_empty, sup_step];

val Lattice_cs = Lattice1_cs addIs [meet_isoton1, meet_isoton2,
				    join_isoton1, join_isoton2, 
				    cmpl_antiton, inf_antiton, sup_isoton];


val AtomLA_ss = AtomLA2_ss;
val AtomL_css = AtomL2_css;
 
val lpre_tac    = pre_tac Lattice_ss AtomLA_ss;
val lstep_tac   = step_tac (op addss2 AtomL_css);
val lfast_tac   = fast_tac (op addss2 AtomL_css);
val lslow_tac   = slow_tac (op addss2 AtomL_css);
fun lpg     str = pg str [lpre_tac 1, REPEAT (FIRSTGOAL lfast_tac)];
fun lslowpg str = pg str [lpre_tac 1, REPEAT (FIRSTGOAL lslow_tac)];



