(* Name     : Relations.ML - theorems for Relation.thy
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 24-Jun-94
   updated  : 30-Jun-94 sample proofs (dedekind_rule) with incidence
   updated  : 07-Jul-94 proof for comp_isoton
   updated  : 18-Jul-94 proofs for atomization of 
   updated  : 27-Jul-94 proofs for pre-associativity
   updated  : 29-Jul-94 proofs for pre-identity
   updated  : 08-Aug-94 using destruction tactic for 2 premises: datac
   updated  : 17-Aug-94 using datac for handling associativity in pre_tac
   updated  : 18-Aug-94 pre2_tac_
   updated  : 09-Nov-94 mtac
   updated  : 16-Dec-94 schr_tac
   updated  : 21-Dec-94 assoc_tac
   updated  : 27-Dec-94 ident_tac
   updated  : 11-Feb-97 adaptions for new HOL version
*)

open Relations;

val  g =  g_ Relations.thy;
val pg = pg_ Relations.thy;
fun lpg' str = pg str [lpre_tac 1, 
		REPEAT (fast_tac (op addss2 AtomL_css addDs [atom_in_atom]
						      addIs [atom_in_refl])1)];

(**** for Composition ********************************************************)

val schroeder_1a = schroeder1 RS iffD1;
val schroeder_1b = schroeder1 RS iffD2;
val schroeder_2a = schroeder2 RS iffD1;
val schroeder_2b = schroeder2 RS iffD2;

val comp_isoton1 = pg "PQ ==> PRQR" [
				rtac schroeder_2b 1,
				rtac incl_trans 1,
				etac (cmpl_antiton RS iffD1) 2,
				rtac schroeder_2a 1,
				rtac incl_refl 1];
val comp_isoton2 = pg "RS ==> PRPS" [
				rtac schroeder_1b 1,
				rtac incl_trans 1,
				etac (cmpl_antiton RS iffD1) 2,
				rtac schroeder_1a 1,
				rtac incl_refl 1];
val comp_isoton1_a = pg "[| aPS; !a. aP --> aQ |] ==> aQS" [
				rtac (atom_in_incl RS iffD1 RS allE) 1,
				mp_tac 2,
				atac 2,
				dtac (atom_in_incl RS iffD2) 1,
				etac comp_isoton1 1];
val comp_isoton2_a = pg "[| aPR; !a. aR --> aS |] ==> aPS" [
				rtac (atom_in_incl RS iffD1 RS allE) 1,
				mp_tac 2,
				atac 2,
				dtac (atom_in_incl RS iffD2) 1,
				etac comp_isoton2 1];
				
val comp_hcontinous1= pg "Inf(A)Q  Inf({RQ | R. RA})" [
		 		fast_tac (AtomL_cs2 addEs [comp_isoton1_a]) 1];
val comp_hcontinous2= pg "QInf(A)  Inf({QR | R. RA})" [
				fast_tac(AtomL_cs2 addEs [comp_isoton2_a]) 1];
val comp_hcontinous12 = pg "Inf({f(R) | R. RA})Q  Inf({f(R)Q | R. RA})" [
				fast_tac (AtomL_cs2 addEs [comp_isoton1_a]) 1];
val comp_hcontinous22 = pg "QInf({f(R) | R. RA})  Inf({Qf(R) | R. RA})" [
				fast_tac (AtomL_cs2 addEs [comp_isoton2_a]) 1];

val comp_continous1= pg "Sup(A)Q = Sup({RQ | R. RA})" [
				rtac incl_antisym 1,
				rtac schroeder_2b 1,
				simp_tac Lattice_ss 1,
				rtac (inf_glb RS mp) 1,
				safe_tac set_cs,
				mon_tac comp_hcontinous1 1,
				mon_tac inf_lb1 1,
				fast_tac Lattice_cs 1,
				fast_tac (Lattice_cs addIs [schroeder_2a]) 1,
				fast_tac (AtomL_cs2 addEs [comp_isoton1_a]) 1];
val comp_continous2= pg "QSup(A) = Sup({QR | R. RA})" [
				rtac incl_antisym 1,
				rtac schroeder_1b 1,
				simp_tac Lattice_ss 1,
				rtac (inf_glb RS mp) 1,
				safe_tac set_cs,
				mon_tac comp_hcontinous2 1,
				mon_tac inf_lb1 1,
				fast_tac Lattice_cs 1,
				fast_tac (Lattice_cs addIs [schroeder_1a]) 1,
				fast_tac (AtomL_cs2 addEs [comp_isoton2_a]) 1];

val atom_in_comp_lemma = pg 
"qSup({Sup({rs|r. r{r.rR}})|s. s{s.sS}}) ==> (r s. rR & sS & qrs)" [
				rewtac incidence_def,
				fast_tac AtomL_cs2 1];
val atom_in_comp_ = pg "qRS = (r s. rR & sS & qrs)" [
				safe_tac HOL_cs,
				rtac atom_in_comp_lemma 1,
				simp_tac (HOL_ss addsimps 
					[comp_continous1 RS sym,
					 comp_continous2 RS sym,
					 strong_atomic]) 1,
				rewtac incidence_def,
				safe_tac HOL_cs,
				rtac comp_isoton1_a 1,
				etac comp_isoton2_a 1,
				fast_tac (HOL_cs addDs [atom_in_atom]) 1,
				fast_tac (HOL_cs addDs [atom_in_atom]) 1];
val atom_in_comp = pg "qRS = (r s. qrs & rR & sS)" [
				sstac atom_in_comp_ 1,
				fast_tac HOL_cs 1];
val atom_in_comp__ = pg "qRS = (r. rR & (s. sS & qrs))" [
				sstac atom_in_comp_ 1,
				fast_tac HOL_cs 1];


(**** for Conversion *********************************************************)

val conv_eq = pg "(R = R^) = (R^ = R)" [ rtac iffI 1, etac sym 1, etac sym 1];

val conv_invol = pg "R^^ = R" [
				rtac incl_antisym 1,
				s_tac (*"R^^"*) I_def2 1,
				s_tac cmpl_invol 1,
				s_tac (*"I"*) cmpl_invol 1,
				rtac schroeder_1a 1,
				rtac schroeder_1a 1,
				simp_tac (Lattice_ss addsimps [I_def2]) 1,
				s_tac (*"R"*) I_def2 1,
				rtac schroeder_1b 1,
				rtac schroeder_1b 1,
				simp_tac (Lattice_ss addsimps [I_def2]) 1];

val conv_isoton2 = pg "RS ==> R^S^" [
				s_tac (*"R^"*) I_def2 1,
				s_tac cmpl_invol 1,
				s_tac (*"I"*) cmpl_invol 1,
				rtac schroeder_1a 1,
				rtac schroeder_2b 1,
				sstac cmpl_invol 1,
				dtac (cmpl_antiton RS iffD1) 1,
				ra2tac incl_trans 1,
				s_tac (*"I"*) cmpl_invol 1,
				rtac schroeder_2a 1,
				rtac schroeder_1b 1,
				simp_tac (Lattice_ss addsimps [I_def2]) 1];
val conv_isoton1 = pg "R^S^ ==> RS" [
				s_tac conv_invol 1,
				s_tac (*"R"*) conv_invol 1,
				ratac conv_isoton2 1];
val conv_isoton = pg "R^S^ = RS" [fast_tac (HOL_cs addEs [conv_isoton1,
				conv_isoton2]) 1];
val conv_equal = pg "(R^=S^) = (R=S)" [
				fast_tac (Lattice_cs addEs [conv_isoton1]) 1];

val conv_O_lemma = pg "x^=O ==> x=O" [
				sstac equal_O 1,
				s_tac (*"x"*) conv_invol 1, (*1*back();*)
				etac ssubst 1,
				s_tac conv_invol 1,
				rtac conv_isoton2 1,
				rtac O_def 1];
val conv_atom2 = pg "Atomic(a) ==> Atomic(a^)" [
				rewtac atom_def,
				safe_tac HOL_cs,
				dtac conv_O_lemma 1,
				contr_tac 1,
				step_tac HOL_cs 1,
				step_tac HOL_cs 1,
				rtac (conv_invol RS sym) 3,
				s_tac conv_invol 2,
				etac conv_isoton2 2,
				dtac conv_O_lemma 1,
				contr_tac 1];
val conv_atom1 = pg "Atomic(a^) ==> Atomic(a)" [
				s_tac conv_invol 1,
				etac conv_atom2 1];
val atomlemma_cs = HOL_cs addEs [conv_atom2];
val conv_atom = pg "Atomic(a^) = Atomic(a)" [
				fast_tac (atomlemma_cs addDs [conv_atom1]) 1];

val conv_atom3 = pg "a^R ==> Atomic(a)" [
				rewtac atom_in_def,
				fast_tac (HOL_cs addEs [conv_atom1]) 1];
val conv_atom4 = pg "aR ==> Atomic(a^)" [
				rewtac atom_in_def,
				fast_tac atomlemma_cs 1];

val atom_in_conv = pg "tR^  = t^R" [
				step_tac AtomL_cs2 1,
				s_tac conv_invol 1,
				subst_tac "t" conv_invol 2, (*2*back();*)
				rewtac atom_in_def,
				fast_tac (atomlemma_cs addEs [conv_isoton2]) 1,
				fast_tac (atomlemma_cs addEs [conv_isoton2])1];

val incidence_is_atoms = map 
	(fn s => pg s [rewtac incidence_def, fast_tac (fst AtomL_css) 1])
	["qrs ==> Atomic(q)", "qrs ==> Atomic(r)", "qrs ==> Atomic(s)"];
val [incidence_is_atom0, incidence_is_atom1, incidence_is_atom2] = 
     incidence_is_atoms;
val incidence_is_atoms_t = incidence_is_atoms @ 
			  (incidence_is_atoms RL [conv_atom1]) @
			  (incidence_is_atoms RL [conv_atom2]);

(* mtac is a kind of substitution tactic like s_tac/sstac, but uses
   an inclusion rule and reasons by the monotonicity of ,,,^,- *)
val mtac_ = let
		val const_left  = pg "OO  x ==> OO  x" [atac 1];
		val const_right = pg "x  OO ==> x  OO" [atac 1];
	    in mtac__ (incl_trans,const_left,const_right) end;
val Relation_ms = 
       [meet_isoton1, meet_isoton2, join_isoton1, join_isoton2, conv_isoton2,
	comp_isoton1, comp_isoton2, cmpl_antiton RS iffD1];
fun mtac thm = mtac_ Relation_ms (rtac thm);
(* like mtac, but the left side of thm  is instantiated by str *)
fun m1tac str thm = let val mon = pg "s_st_t ==> s_st_t" [atac 1] in
		mtac_ Relation_ms (res2_inst_tac thm [("s_s",str)] mon) end;
(* like mtac, but the right side of thm  is instantiated by str *)
fun m2tac str thm = let val mon = pg "s_st_t ==> s_st_t" [atac 1] in
		mtac_ Relation_ms (res2_inst_tac thm [("t_t",str)] mon) end;
val matac = mtac_ Relation_ms atac;



(**** for Schroeder equivalences *********************************************)

val schr_lemma1a = lpg' "[| q  rs; Atomic(q) |] ==> ~(r--s  -q)";
val schr_lemma1b = lpg' "[| ~(r--s  -q); Atomic(q) |] ==> q  rs";
val schr_1 = pg "sr^q = qrs" [
			rewrite_goals_tac [incidence_def, atom_in_def],
			step_tac (HOL_cs addSDs [schr_lemma1a,conv_atom1]
					 addSIs [schr_lemma1b,conv_atom2]) 1,
			etac swap 1,
			rtac schroeder_1a 1,
			asm_full_simp_tac Lattice_ss 1,
			etac swap 1,
			rtac schroeder_1b 1,
			asm_full_simp_tac Lattice_ss 1];

val schr_lemma2a = lpg' "[| q  rs; Atomic(q) |] ==> ~(--rs  -q)";
val schr_lemma2b = lpg' "[| ~(--rs  -q); Atomic(q) |] ==> q  rs";
val schr_2 = pg "rqs^ = qrs" [
			rewrite_goals_tac [incidence_def, atom_in_def],
			step_tac (HOL_cs addSDs [schr_lemma2a,conv_atom1]
					 addSIs [schr_lemma2b,conv_atom2]) 1,
			etac swap 1,
			rtac schroeder_2a 1,
			asm_full_simp_tac Lattice_ss 1,
			etac swap 1,
			rtac schroeder_2b 1,
			asm_full_simp_tac Lattice_ss 1];

val schr_0a  = pg "q^sr   = qr^s^" [
				sstac schr_1 1,
				s_tac schr_2 1,
				rtac (schr_1 RS sym) 1];

val schr_ss = HOL_ss addsimps [schr_1, schr_2];
val schr_0  = pg "q^s^r^ = qrs" [simp_tac schr_ss 1];
val schr_1a = pg "s^q^r  = qrs" [simp_tac schr_ss 1];
val schr_2a = pg "r^s q^ = qrs" [simp_tac schr_ss 1];
val Relation1_ss = Lattice_ss addsimps [comp_assoc, I_def1, I_def2,
					conv_invol, conv_isoton];

use "Relations_tac1";


val AtomR1_Es = [conv_atom1, conv_atom2,
		 conv_atom3, conv_atom4] @ 
		 incidence_is_atoms_t;
val AtomRA1_ss = AtomLA_ss addsimps [atom_in_comp, atom_in_conv];
val AtomR1_css = curry (pair_upd1 op addaltern)
		(AtomL_css addsimps2 [(*conv_eq, *)conv_equal,
				      conv_invol, conv_atom]
			(*addsimps2 ([conv_atom, 
				      conv_atom3, conv_atom4] @ 
				     incidence_is_atoms_t)*)
			   addSEs2' ([conv_atom1, conv_atom2,
				      conv_atom3, conv_atom4] @ 
			 	     incidence_is_atoms_t))
		(FIRST' [atom_tac, atomin_tac, schr_tac]);

val r1pre_tac  = pre_tac Relation1_ss AtomRA1_ss;
val r1slow_step_tac = slow_step_tac (op addss2 AtomR1_css);
val r1step_tac = step_tac (op addss2 AtomR1_css);
val r1fast_tac = fast_tac (op addss2 AtomR1_css);
fun r1pg str   = pg str [r1pre_tac 1, REPEAT (FIRSTGOAL r1fast_tac)];

(**** for Composition, 2 *****************************************************)

val comp_O1 = r1pg "OR = O";
val comp_O2 = r1pg "RO = O";

val comp_isoton = r1pg "[| P  Q; R  S |] ==> PR  QS";

val comp_hdistr_meet1 = r1pg "Q(RS)  QRQS";
val comp_hdistr_meet2 = r1pg "(RS)Q  RQSQ";

val comp_distr_join1 = r1pg "Q(RS) = QRQS";
val comp_distr_join2 = r1pg "(RS)Q = RQSQ";


(**** for Conversion, 2 ******************************************************)

val conv_comp = r1pg "(RS)^  =  S^R^";
val conv_cmpl = r1pg "-R^ = -(R^)";

val conv_join = r1pg "(RS)^  =  S^R^";
val conv_meet = r1pg "(RS)^  =  S^R^";

val conv_sup = r1pg "(Sup(A))^  =  Sup({R^ | R. RA})";
val conv_inf = r1pg "(Inf(A))^  =  Inf({R^ | R. RA})";

val conv_O = r1pg "O^ = O";
val conv_L = r1pg "L^ = L";

val dedekind_rule = r1pg "QRS  (QSR^)(RQ^S)";
val dedekind_1a   = r1pg "QRS  Q(RQ^S)";
val dedekind_1b   = r1pg "SQR  Q(RQ^S)";
val dedekind_2a   = r1pg "QRS  (QSR^)R";
val dedekind_2b   = r1pg "SQR  (QSR^)R";
val schroe_1 = r1pg "(QRS = O) = (QSR^ = O)";
val schroe_2 = r1pg "(QRS = O) = (Q^SR = O)";


(**** misc *******************************************************************)

val luce1 = pg "[| XS; QLX; QX^R |] ==> QRS" [
				rtac incl_trans 1,
				etac meet_glb 1,
				rtac incl_refl 1,
				mon_tac dedekind_rule 1,
				r1pre_tac 1,
				r1fast_tac 1];
val luce2 = pg "[| XR; QXL; X^QS |] ==> QRS" [
				rtac incl_trans 1,
				etac meet_glb 1,
				rtac incl_refl 1,
				mon_tac dedekind_rule 1,
				pre_tac (Relation1_ss delsimps [comp_assoc])
					AtomRA1_ss 1,
				r1fast_tac 1];

val L_unequal_O_lemma = r1pg "R=O ==> RLR=R";
val L_unequal_O = pg "(L::('a,'b)rel)  O" [
				rtac ccontr 1,
				dtac notnotD 1,
				double_prem_tac 1,
				dtac L_unequal_O_lemma 1,
				dtac (tarski_rule RS iffD2) 1,
				etac swap 1,
				s_tac conv_equal 1,
				sstac conv_L 1,
				sstac conv_O 1,
				atac 1];

val I_unequal_O_lemma = r1pg "R=O --> RL=O";
val I_unequal_O = pg "I  O" [
				rtac (I_unequal_O_lemma RS mb) 1,
				sstac I_def1 1,
				rtac L_unequal_O 1];

val L_is_vector = pg "LL=L" [
				s_tac I_def2 1, (* 3*back(); *)
				s_tac tarski_rule 1,
				rtac I_unequal_O 1];

val O_nulls_L_lemma = r1pg "R=O ==> LR=O";
val O_nulls_L_lemma1 = prove_goal thy "R=O ==> R(L::('a,'b)rel)" (fn prems => [
				simp_tac (HOL_ss addsimps prems) 1,
				rtac notI 1,
				dtac sym 1,
				cut_facts_tac [L_unequal_O] 1,
				etac swap2 1,
				rtac L_unequal_O 1]);
val only_O_nulls_L = pg "(RL=O) = (R=O)" [
				rtac iffI 1,
				etac swap2 1,
				dtac (tarski_rule RS iffD1) 1,
				etac swap2 1,
				dtac notnotD 1,
				dtac O_nulls_L_lemma 1,
				s_tac comp_assoc 1,
				datac O_nulls_L_lemma1 1,
				etac ssubst 1,
				rtac comp_O1 1];


(**** for Identity ***********************************************************)

val trace_subst = pg "[| P (i s); i t =  i s |] ==> P (i t)" [
				asm_simp_tac HOL_ss 1];

val conv_I_lemma = pg "I^  I" [
				subst_tac "I^" I_def2 1,
				s_tac cmpl_invol 1, (*14*back();*)
				s_tac schroeder1 1,
				simp_tac Relation1_ss 1];
val conv_I = pg "I^ = I" [
				rtac incl_antisym 1,
				rtac conv_I_lemma 1,
				rtac conv_isoton1 1,
				sstac conv_invol 1,
				rtac conv_I_lemma 1];

val Relation2_ss = Relation1_ss addsimps [comp_O1, comp_O2, conv_comp,
				  conv_cmpl,
				  conv_O, conv_L, conv_I, conv_equal,
				  conv_meet, conv_join, conv_inf, conv_sup,
				  L_unequal_O, I_unequal_O, L_is_vector];


val atom_conv_in_I = pg "a^I = aI" [
				s_tac atom_in_conv 1,
				simp_tac Relation2_ss 1];

val trace_in_I = pg "i(j)  I" [
				rewtac trace_def,
				fast_tac AtomL_cs2 1];

val conv_trace = pg "i(j)^ = i(j)" [
				rewtac trace_def,
				simp_tac Relation2_ss 1];
val conv_trace_eq = pg "(i a = b^) = (i a = b)" [
				s_tac conv_equal 1,
				stac conv_trace 1,
				rtac refl 1];

local
val trace_is_atom_le = pg "!!a. [| Atomic(a); aRa  O |] ==> a  RL"
		[r1pre_tac 1,
		 REPEAT (fast_tac (op addss2 AtomR1_css addDs[atom_in_atom])1)];
val trace_is_atom_lem =   pg "-(R^)R  -I" [
				s_tac conv_cmpl 1,
				rtac (schroeder1 RS iffD2) 1,
				simp_tac Relation2_ss 1];
val trace_is_atom_lemm = lpg' "I-(R^)(IR-(R^)R)  I-(R^)R";
val trace_is_atom_lemma = pg "(I-R)L  -((IR)L)" [
				rtac (schroeder1 RS iffD2) 1,
				simp_tac Relation2_ss 1,
				mtac comp_hdistr_meet1 1,
				sstac I_def2 1,
				mtac meet_isoton2 1,
				rtac comp_hdistr_meet2 1,
				mtac trace_is_atom_lemm 1,
				mtac meet_isoton2 1,
				rtac trace_is_atom_lem 1,
				simp_tac Relation2_ss 1];
in
val trace_is_atom = pg "Atomic(a) ==> Atomic(i(a))" [
				rewrite_goals_tac [atom_def, trace_def],
				step_tac HOL_cs 1,
				dtac (schroe_2 RS iffD2) 1,
				asm_full_simp_tac Relation2_ss 1,
				forward_tac [incl_antisym] 1,
				atac 2,
				mtac comp_isoton2 1,
				rtac L_def 1,
				rtac (incl_equivO RS iffD2) 1,
				sstac meet_assoc 1,
				rtac incl_intro1 1,
				rtac (schroe_1 RS iffD2) 1,
				simp_tac Relation2_ss 1,
				mtac trace_is_atom_lemma 1,
				rtac (incl_equivO RS iffD1) 1,
				rtac trace_is_atom_le 1,
				rtac conv_atom2 1,
				rewtac atom_def,
				fast_tac HOL_cs 1,
				s_tac schroe_1 1,
				s_tac meet_assoc 1,
				fold_goals_tac [trace_def],
				rotate_tac 2 1, etac swap 1,
				step_tac HOL_cs 1,
				step_tac HOL_cs 1,
				step_tac HOL_cs 1,
				fast_tac HOL_cs 1,
				rtac incl_refl 1,
				etac swap 1,
				lpre_tac 1,
				lfast_tac 1];
end;

val trace_idempot' = pg "Atomic a ==> i(i a) = i(a)" [
				rtac atom_in_atom 1,
				rewtac atom_in_def,
				safe_tac (HOL_cs addSIs [trace_is_atom]),
				rewtac trace_def,
				mtac meet_lb2 1,
				simp_tac Relation2_ss 1,
				rtac meet_lb2 1];

val trace_of_conv = pg "i(a^) = aa^I" [
				rewtac trace_def,
				simp_tac Relation2_ss 1];
				
val atom_in_I = pg "jI = (Atomic(j) & (i(j) = j))" [
				rewrite_goals_tac [trace_def, atom_in_def],
				step_tac HOL_cs 1,
				rtac atom_in_atom 1,
				rewtac atom_in_def,
				step_tac HOL_cs 1,
				mtac comp_isoton1 2,
				etac conv_isoton2 2,
				simp_tac Relation2_ss 2,
				simp_tac Relation2_ss 2,
				fast_tac AtomL_cs2 2,
				etac subst 2,
				fast_tac AtomL_cs2 2,
				fold_goals_tac [trace_def],
				etac trace_is_atom 1];

val sym_lemma = pg "RI ==> RR^" [
				s_tac I_def1 1, (*3*back();*)
				mon_tac meet_glb 1,
				rtac incl_refl 1,
				sstac I_def1 1,
				atac 1,
				mon_tac dedekind_rule 1,
				m1tac "RI^I" meet_lb2 1,
				mtac meet_lb2 1,
				simp_tac Relation2_ss 1];
val trace_sym = pg "[| i a = b; Atomic b |] ==> b^ = b" [
				rtac (atom_in_atom RS sym) 1,
				 etac conv_atom2 2,
				rewtac atom_in_def,
				safe_tac HOL_cs,
				rtac (trace_in_I RS sym_lemma) 1];
val I_D = pg "jI ==> i(j)=j & j^=j & Atomic(j)" [
		dtac (atom_in_I RS iffD1) 1,
		fast_tac (HOL_cs addSIs [trace_sym]) 1];
val I_coreflD = I_D RS conjunct2 RS conjunct1;

val I_I1 = pg "Atomic(j) ==> i(j )I" [
			rewtac atom_in_def,
			fast_tac (HOL_cs addIs [trace_is_atom,trace_in_I]) 1];
val I_I2 = pg "Atomic(j) ==> i(j^)I" [etac (conv_atom2 RS I_I1) 1];

local
val ident_lemma2 = pg "(aaI^)(Ia^a) = a(a^aI)" [
				simp_tac Relation2_ss 1];
in
val ident_I2_ = pg "rri r = Atomic r" [
				rewrite_goals_tac [trace_def, incidence_def,
						   atom_in_def],
				step_tac HOL_cs 1,
				s_tac ident_lemma2 1,
				rtac incl_trans 1,
				rtac dedekind_rule 2,
				sstac I_def2 1,
				fast_tac AtomL_cs2 1,
				fold_goals_tac [trace_def],
				etac trace_is_atom 1];
end;

val ident_I0_ = pg "i rr^r = Atomic r" 
			[simp_tac (Relation2_ss addsimps [schr_1,ident_I2_]) 1];
val ident_I0t_ = pg "i(r^)rr^ = Atomic r" [
				s_tac conv_invol 1, (*?*back();*)
				s_tac conv_atom 1,
				rtac ident_I0_ 1];
val ident_I1t_ = pg "r^i rr^ = Atomic r" 
			[simp_tac (Relation2_ss addsimps [schr_2,ident_I0_]) 1];
val ident_I1_ = pg "ri(r^)r = Atomic r" [
				s_tac conv_invol 1, (*7*back();*)
				s_tac conv_atom 1,
				rtac ident_I1t_ 1];

val ident_Ii = pg "Atomic r ==> i ri ri r" [
				rtac (trace_idempot' RS subst) 1,
				 atac 1,
				rtac (ident_I2_ RS iffD2) 1,
				etac trace_is_atom 1];
				

val ident_D2__ = pg "qri(j) ==> q=r" [
				rtac atom_in_atom 1,
				rewrite_goals_tac [trace_def, incidence_def,
							atom_in_def],
				step_tac HOL_cs 1,
				ratac incl_trans 1,
				mtac meet_lb2 1,
				simp_tac Relation2_ss 1];


val ident_eq2 = pg "Atomic j ==> qri(j) = (q=r  i j=i r  Atomic r)" [
				safe_tac (HOL_cs addSEs [ident_D2__]),
				  etac incidence_is_atom1 2,
				 etac ssubst 2,
				 etac (ident_I2_ RS iffD2) 2,
				forward_tac [ident_D2__] 1,
				hyp_subst_tac 1,
				dtac (schr_1 RS iffD2) 1,
				rtac atom_in_atom 1,
				 etac (incidence_is_atom2 RS trace_is_atom) 2,
				dtac I_I1 1,
				rewrite_goals_tac [incidence_def, trace_def],
				fast_tac(HOL_cs addIs[atom_in_meet RS iffD2])1];

val ident_eq0 = pg "Atomic j ==> i(j)rq = (q=r^  i j=i(r^)  Atomic r)" [
				s_tac schr_1 1,
				asm_simp_tac (HOL_ss addsimps 
					[ident_eq2,conv_atom]) 1];
val ident_eq1 = pg "Atomic j ==> ri(j)q = (r=q  i j=i(q^)  Atomic q)" [
				s_tac schr_2 1,
				asm_simp_tac (HOL_ss addsimps
					[ident_eq0,conv_equal]) 1,
				fast_tac HOL_cs 1];

val ident_D2 = pg "[| qki(j); Atomic j |] ==> q=k   i j = i k    Atomic k" [
			fast_tac (HOL_cs addSDs [ident_eq2 RS iffD1])1];
val ident_D0 = pg "[| i(j)kq; Atomic j |] ==> q=k^  i j = i(k^)  Atomic k" [
				fast_tac (HOL_cs addSDs [ident_eq0 RS iffD1])1];
val ident_D1 = pg "[| qi(j)k; Atomic j |] ==> q=k   i j = i(k^)  Atomic k" [
				fast_tac (HOL_cs addSDs [ident_eq1 RS iffD1])1];

val ident_I2 = pg "[| i j = i k;   Atomic k; Atomic j |] ==> kki(j)" [
				fast_tac (HOL_cs addSEs [ident_eq2 RS iffD2])1];
val ident_I0 = pg "[| i j = i(k^); Atomic k; Atomic j |] ==> i(j)kk^" [
				fast_tac (HOL_cs addSEs [ident_eq0 RS iffD2])1];
val ident_I0t= pg "[| i j = i k;   Atomic k; Atomic j |] ==> i(j)k^k" [
				fast_tac (HOL_cs addIs [conv_invol RS ssubst]
				      addSEs [conv_atom2,ident_eq0 RS iffD2])1];
val ident_I1 = pg "[| i j = i(k^); Atomic k; Atomic(j) |] ==> ki(j)k" [
				fast_tac (HOL_cs addSEs [ident_eq1 RS iffD2])1];

local
val lemma = pg "[| i q=i r; Atomic q; Atomic r |] ==> h. qhr  hqr^" [
				subgoal_tac "q(qr^)r" 1,
				 dtac (atom_in_comp_ RS iffD1) 1,
				  safe_tac HOL_cs,
				 datac atom_in_atom 1,
				 rewtac incidence_def,
				 fast_tac (HOL_cs addEs [conv_atom2,
						atom_in_is_atom]) 1,
				rewtac atom_in_def,
				ratac conjI 1,
				dtac (ident_I2_ RS iffD2) 1,
				s_tac comp_assoc 1,
				asm_full_simp_tac (HOL_ss addsimps 
				    [trace_def,incidence_def, atom_in_def])1,
				fast_tac (HOL_cs addEs [incl_trans]
					addIs [comp_isoton2,meet_lb1]) 1];
in
val trace_trace_D = pg "[| i q=i r; Atomic q; Atomic r |] ==> h. qhr" [
				fast_tac (HOL_cs addSDs [lemma]) 1];
end;


(**** for Associativity ******************************************************)

val atom_in_comp_lemm = 
pg "[| a  RS; Atomic(a) |] ==> (r s. rR & sS & ars)" [
				rtac (atom_in_comp_ RS iffD1) 1,
			asm_full_simp_tac (HOL_ss addsimps [atom_in_def]) 1];
val comp_atom_assoc20 = pg "[| exd; dyz |] ==> (f. fxy & efz)" [
				rewrite_goals_tac [incidence_def, atom_in_def],
				step_tac HOL_cs 1,
				dtac comp_isoton2 1, (*1*back();*)
				dtac incl_trans 1,
				asm_full_simp_tac Relation2_ss 1,
				datac atom_in_comp_lemm 1,
				safe_tac HOL_cs,
				datac atom_in_atom 1,
				rewrite_goals_tac [incidence_def, atom_in_def],
				fast_tac HOL_cs 1];
val comp_atom_assoc02 = pg "[| dyz; exd |] ==> (f. fxy & efz)" [
				datac comp_atom_assoc20 1,
				atac 1];
val comp_atom_assoc00 = pg "[| dxe; dyz |] ==> (f. fx^y & efz)" [
				dtac (schr_1 RS iffD2) 1,
				datac comp_atom_assoc20 1,
				atac 1];
val comp_atom_assoc22 = pg "[| exd; zyd |] ==> (f. fxy^ & efz)" [
				dtac (schr_1 RS iffD2) 1, (*1*back();*)
				datac comp_atom_assoc20 1,
				atac 1];

val comp_atom_assoc01 = pg "[| dxy; edz |] ==> (f. exf & fyz)" [
	(* by ^-Duality *)	dtac (schr_0 RS iffD2) 1,
				dtac (schr_0 RS iffD2) 1,
				datac comp_atom_assoc20 1,
				safe_tac HOL_cs,
				step_tac HOL_cs 1,
				step_tac HOL_cs 1,
				etac swap2 1, (*2*back();*)
				s_tac conv_invol 1, (*2*back();*)
				etac swap 1,
				dtac notnotD 1,
				etac (schr_0 RS iffD1) 1,
				etac swap2 1, (*back();*)
				s_tac conv_invol 1, (*4*back();*)
				etac swap 1,
				dtac notnotD 1,
				etac (schr_0 RS iffD1) 1];
val comp_atom_assoc10 = pg "[| edz; dxy |] ==> (f. exf & fyz)" [
				datac comp_atom_assoc01 1,
				atac 1];
val comp_atom_assoc11 = pg "[| xdy; edz |] ==> (f. exf & fy^z)" [
				dtac (schr_2 RS iffD2) 1,
				datac comp_atom_assoc01 1,
				atac 1];
val comp_atom_assoc12 = pg "[| edz; yxd |] ==> (f. ex^f & fyz)" [
				dtac (schr_1 RS iffD2) 1, (*1*back();*)
				datac comp_atom_assoc10 1,
				atac 1];
val comp_atom_assoc21 = pg "[| yxd; edz |] ==> (f. ex^f & fyz)" [
				datac comp_atom_assoc12 1,
				atac 1];

use "Relations_tac2";

val AtomRA2_ss = AtomRA1_ss;
val AtomR2_css = curry (pair_upd1 op addaltern)
		(curry (pair_upd1 op addSaltern)
		(AtomR1_css addsimps2 [atom_conv_in_I,conv_trace]
			    addSEs2 [trace_subst]
			(*addsimps2 [I_I1, I_I2,trace_is_atom]*)
			    addSIs2 [I_I1, I_I2,trace_is_atom]
			(*addsimps2 [ident_Ii, ident_I0, ident_I0t,
				     ident_I1, ident_I2]*)
			    addSIs2 [ident_Ii, ident_I0, ident_I0t, 
				     ident_I1, ident_I2]
			    addSDs2 [ident_D0, ident_D1, ident_D2])
		(FIRST' [ident_tac, assoc_tac]))
		(fn i => dtac trace_trace_D i (*THEN let
				val trace_trace_D_var_tests = [
				 pg "(h. Ohr) ==> True" [rtac TrueI 1],
				 pg "(h. rhO) ==> True" [rtac TrueI 1]];
			in IF (dresolve_tac trace_trace_D_var_tests (i+2)) 
			   no_tac all_tac end*));

val r2pre_tac  = pre_tac Relation2_ss AtomRA2_ss;
val r2slow_step_tac = slow_step_tac (op addss2 AtomR2_css);
val r2step_tac = step_tac (op addss2 AtomR2_css);
val r2fast_tac = fast_tac (op addss2 AtomR2_css);
fun r2pg str   = pg str [r2pre_tac 1, REPEAT (FIRSTGOAL r2fast_tac)];


(*****************************************************************************)

val idempotI = r2pg "RI ==> RR=R";
val trace_idempot = pg "i(i a) = i(a)" [
				rewtac trace_def,
				stac (r2pg "RI ==> R^=R") 1,
				 rtac meet_lb2 1,
				stac idempotI 1,
				 rtac meet_lb2 1,
				rtac (r2pg "x  y ==> xy = x") 1,
				rtac meet_lb2 1];

(*
fun NEXT_GOAL p tac i = let
  fun find_goal j (Const("==>",_)$t$u) = if j<0 then 1+find_goal (j+1) u
					 else if p t then j 
					      else find_goal (j+1) u
  |   find_goal j _                    = 0;
in fn s => s |> (tac (1 + find_goal (1-i) (#prop(rep_thm s)))) end;
fun is_var_atom (Const("all",_)$Abs(_,_,t)) = is_var_atom t
|   is_var_atom (Const("==>",_)$
	(Const("Trueprop",_)$(Const("not",_)$(Const("Atomic",_)$u)))$_)
							= is_var u
|   is_var_atom (Const("==>",_)$_$u)			= is_var_atom u
|   is_var_atom (Const("Trueprop",_)$(Const("Atomic",_)$u))	= is_var u
|   is_var_atom _						= false;
*)

val Relation_ss = Relation2_ss;
val Relation_cs = Lattice_cs addIs [comp_isoton1, comp_isoton2, conv_isoton2];


val AtomRA_ss = AtomRA2_ss delsimps [atom_in_comp] addsimps [atom_in_comp__];
val AtomR_css = (*curry (pair_upd1 op compWrapper)*)
		(AtomR2_css addsimps2 [trace_idempot])
		(*(NEXT_GOAL (not o is_var_atom))*);
val rpre_tac    = pre_tac Relation_ss AtomRA_ss;
val rslow_step_tac = slow_step_tac (op addss2 AtomR_css);
val rstep_tac   = step_tac (op addss2 AtomR_css);
val rfast_tac   = fast_tac (op addss2 AtomR_css);
val rslow_tac   = slow_tac (op addss2 AtomR_css);
fun rpg     str = pg str [rpre_tac 1, REPEAT (FIRSTGOAL rfast_tac)];
fun rslowpg str = pg str [rpre_tac 1, REPEAT (FIRSTGOAL rslow_tac)];

