(* Name     : Relations.thy - Relation Algebraic Language and Logic
	      - an Isabelle theory for RALL -
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 20-May-94
   updated  : 27-May-94 rules
   updated  : 29-Jun-94 incidence and ident
   updated  : 13-Jul-94 merged with Monoidal.thy
   updated  : 11-Feb-97 adaptions for new HOL version
*)

Relations = Lattice +


classes

  relation < lattice


types

  ('a, 'b) rel			(* unfortunately restricted to rel, because *)
				(* type classes cannot have parameters      *)

arities

  rel :: (term,term)relation


consts

  (* Composition and Identity *)
  Comp	:: "[('a,'b)rel,('b,'c)rel] => ('a,'c)rel"	("(_/_)"   [80,81] 80)
  I	:: "('a,'a)rel"

  (* Conversion *)
  Conv  :: " ('a,'b)rel             => ('b,'a)rel"	("(_^)"     [90]    90)

  (* Incidence and Pre-Identity *)
  AtomInComp :: "[('a,'c)rel, ('a,'b)rel ,('b,'c)rel]   \
		 \		    => bool"		("(___)"[52,52,52]51)
  i	::  "('a,'b)rel => ('b,'b)rel"


rules

  (* for Compositon, that is, Multiplication of Relations *)
  comp_assoc	"Q(RS)  =  (QR)S"

  I_def1	"IR  =  R"
  I_def2	"RI  =  R"

  (* Schroeder equivalences *)
  schroeder1	"(QR  S) = (Q^-S  -R)"
  schroeder2	"(QR  S) = (-SR^  -Q)"

  (* Tarski rule *)
  tarski_rule	"R  O  =  (LRL = L)"

  (* Auxiliaries *)
  incidence_def  "axy   == (axy & Atomic(x) & Atomic(y))"
  trace_def	 "i(a)    == a^a  I"


end
