(* Name     : Relations_tac1.ML - tactics for Relations.thy
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1995

   created  : 31-Jan-95, extracted from Relation.ML
*)

fun flatmap f [] = []
  | flatmap f (x::xs) = f x xs @ flatmap f xs;
fun num p []      = ~1
|   num p (x::xs) = if p x then 0 else case num p xs of
		      ~1 => ~1
		    | n  => n+1;
fun numeq x = num (fn z => z=x);
fun nth _ []      = Bound ~1
|   nth n (x::xs) = if n=0 then x else nth (n-1) xs
fun del _ []      = []
|   del z (x::xs) = if x=z then xs else x::(del z xs);
fun is_permut []      l = l=[]
|   is_permut (x::xs) l = numeq x l >= 0 andalso is_permut xs(del x l);

local
fun var_arg (Const("at", _) $ a0)   = var_arg a0
|   var_arg (x as Var   _)	    = Some x
|   var_arg (x as Free  _)	    = Some x
|   var_arg (x as Bound _)	    = Some x
|   var_arg (Const("Conv", t) $ a0) = (case var_arg a0 of
			  None    => None
			| Some va => Some (Const("Conv", t) $ va))
|   var_arg (Const("i", t) $ a0)    = (case var_arg a0 of
			  None    => None
			| Some va => Some (Const("i", t) $ va))
|   var_arg (v $ d)		= ((var_arg d); var_arg v)
|   var_arg _			= None;

fun collect_args argf n (Const("==>", _) $ H $ B) = (case argf H of
			None => collect_args argf (n+1) B
		      | Some (t,ia) => let val (r,x) = collect_args argf (n+1) B
				       in ((n,(t,ia))::r,x) end)
|   collect_args argf n (Const("all",_)$Abs(_,_,t)) = collect_args argf n t
|   collect_args argf _ B = ([],argf B);
fun atom_args (Const("Trueprop", _)$(Const("Atomic", t)$a0)) =
	   (case var_arg a0 of Some va0 => Some (t,[va0]) | _ => None)
|   atom_args _ = None;
fun atomin_args (Const("Trueprop", _)$(Const("AtomIn", t)$a0$a1)) =
	   (case var_arg a0 of Some va0 => Some (t,[va0]) | _ => None)
|   atomin_args _ = None;
fun incid_args f (Const("Trueprop", _)$(Const("AtomInComp", t)$a0$a1$a2)) =
	   (case (var_arg(f a0),var_arg (f a1),var_arg(f a2)) of
		  (Some va0,Some va1,Some va2) => Some (t,[va0,va1,va2])
		| _			       => None)
|   incid_args f _ = None;
fun arg_nc (Const("Conv", _) $ a0) = a0
|   arg_nc a			   = a;
fun collect f s i = collect_args f 0 (#3(dest_state (s, i)));
in
  val list_incids  = collect (incid_args arg_nc);
  val list_incids2 = collect (incid_args I);
  val list_atoms   = collect atom_args;
  val list_atomins = collect atomin_args;
  fun incid_conv(Const("Trueprop",_)$(Const("AtomInComp",_)$a0$a1$a2)) k=
     let val a = nth k [a0,a1,a2] in if arg_nc a=a then 0 else 1 end
  |   incid_conv _ _ = ~1;
end;


fun strip_assums_hyp (Const("==>", _) $ H $ B) = H :: strip_assums_hyp B
  | strip_assums_hyp (Const("all",_)$Abs(a,T,t)) = strip_assums_hyp t
  | strip_assums_hyp B = [];

fun strip_assums_concl (Const("==>", _) $ H $ B) = strip_assums_concl B
  | strip_assums_concl (Const("all",_)$Abs(a,T,t)) = strip_assums_concl t
  | strip_assums_concl B = B;

local
  fun is_conv_var (Const("at"  , _) $ a0       ) = is_conv_var a0
  |   is_conv_var (Const("Conv", _) $ Var(_, _)) = true
  |   is_conv_var _                              = false;
  fun is_no_var_no_conv (Const("at"  ,_) $a0) = is_no_var_no_conv a0
  |   is_no_var_no_conv (Const("Conv",_) $ _) = false
  |   is_no_var_no_conv (Var _)               = false
  |   is_no_var_no_conv  _                    = true;
  fun one_arg ((a,c),rule) = EVERY' (
	if is_conv_var a andalso is_no_var_no_conv c then [rtac (rule RS iffD1)]
   else	if is_conv_var c andalso is_no_var_no_conv a then [dtac (rule RS iffD2),
							   det_rot_tac ~1]
   else []); 
(*solve subgoal i by assumption with the first premise, applying the involution
  of conversion for unification of a list of term pairs like (x,?y^)*)
(*intentionally fails for trivial solution by assumption*)
  fun conv_invol_atac invol_rules (ia,ca) i = DETERM
	(CHANGED (EVERY' (map2 one_arg (ia~~ca,invol_rules)) i) THEN atac i);
  fun prover s = pg s [stac conv_invol 1 THEN rtac refl 1];
  val atom_invols   = map prover ["Atomic (a^^) = Atomic a"];
  val atomin_invols = map prover ["a^^R = aR"];
  val incid_invols  = map prover 
		    ["a^^bc = abc", "ab^^c = abc", "abc^^ = abc"];
  fun could_unify_typ (Type (n1,l1),Type (n2,l2)) = n1=n2 andalso
				forall could_unify_typ (l1~~l2)
  |   could_unify_typ (TFree(n1,s1),TFree(n2,s2)) = n1=n2
  |   could_unify_typ (TVar _      ,_           ) = true
  |   could_unify_typ (_           ,TVar _      ) = true
  |   could_unify_typ (Type _      ,_           ) = false
  |   could_unify_typ (_           ,Type _      ) = false;
  fun maybe_same_type (ct,_) (_,(ht,_)) = could_unify_typ (ct,ht);

in 
(*solve subgoal i by assumption with involution of conversion*)
  fun conv_tac list_args invol_rules tac = let
    fun one_t concl (n,ia) = det_rot_tac n THEN' 
			     conv_invol_atac invol_rules (snd ia,snd concl);
    fun tacf i s = (case list_args s i of (hyps, Some concl) =>
(* hyps  = list of (position_in_prems, (typ,list of (argument))) *)
(* concl = (typ,list of (argument)) *)
        tac i APPEND
	APPEND_ALL' (map (one_t concl) (filter (maybe_same_type concl) hyps)) i 
	                                 | (_,None) => no_tac);
    in (fn i => (fn st => tacf i st st)) end;
  val incid_conv_asm_rl_tac  = conv_invol_atac ;
(*solve 'Atomic'  subgoal i by assumption with involution of conversion*)
(*solve 'AtomIn ' subgoal i by assumption with involution of conversion*)
(*solve incidence subgoal i by assumption with involution of conversion*)
  val atom_conv_tac      = conv_tac list_atoms   atom_invols   (K no_tac);
  val atomin_conv_tac    = conv_tac list_atomins atomin_invols (K no_tac);
  fun incid_conv_tac tac = conv_tac list_incids2 incid_invols  tac;
  val simp_a_incid_conv = DETERM o 
		(match_tac ((incid_invols RL [iffD2])@[asm_rl]));
end;

fun swapify_tac inst tac = let
    val special_swap = read_instantiate_sg (sign_of thy) [("Pa", inst)] swap;
in fn i => tac i APPEND (etac special_swap i THEN tac i) end; 

val atom_tac = swapify_tac "Atomic ?a" atom_conv_tac;

val atomin_tac = swapify_tac "?a?R" atomin_conv_tac;

local
fun schr_tacs ictac = map (fn fs => (DETERM'(EVERY' fs) THEN' ictac atac)) [
  [rtac (schr_1  RS iffD1),simp_a_incid_conv],
  [rtac (schr_2  RS iffD1),simp_a_incid_conv],
  [rtac (schr_1a RS iffD1),simp_a_incid_conv,simp_a_incid_conv],
  [rtac (schr_2a RS iffD1),simp_a_incid_conv,simp_a_incid_conv],
  [rtac (schr_0  RS iffD1),simp_a_incid_conv,simp_a_incid_conv,
					     simp_a_incid_conv]];
in
fun schr_tacf ictac i = let
    fun tacf s = let val (hyps,concl) = list_incids s i in 
		 if concl = None orelse hyps = [] then no_tac else
		 ictac (K no_tac) i APPEND APPEND_ALL' (schr_tacs ictac) i end
  in (fn st => tacf st st) end;
end;
(*solve incidence subgoal i by some schroeder equivalence, 
  but not by trivial assumption or contradiction*)
val schr_tac = swapify_tac "?a?b?c" (schr_tacf incid_conv_tac);

(*solve subgoal i by schroeder equivalence or asm. between the 1st premise and 
  conclusion, with common (up to conversion) non-variable incidence argument k*)
local val conv_thms = [schr_0 RS iffD1, schr_1 RS iffD1, schr_2 RS iffD1];
in
fun conv_atacf ict k = ict atac APPEND' rtac (nth_elem(k,conv_thms)) THEN'
				     (if k>0 then simp_a_incid_conv
	else EVERY'[simp_a_incid_conv, simp_a_incid_conv, simp_a_incid_conv]) 
	     THEN' ict atac;
end;
val conv_atac = conv_atacf incid_conv_tac;


