(* Name     : Special.thy - Special relations
	      - an Isabelle theory for RALL -
   Author   : David von Oheimb, Technische Universitaet Muenchen
   Copyright: 1994

   created  : 03-Aug-94
   updated  : 12-Aug-94 Order
   updated  : 30-Dec-94 Difunctional
*)

Special = Relations +


consts

  (* Functions *)
  Univalent,
  Total,
  Function,
  Injective,
  Surjective,
  Bijective,
  Difunctional	:: "('a,'b)rel => bool"

  (* Order *)
  Reflexive,
  Transitive,
  Preorder,
  Antisymmetric,
  Order	,
  Symmetric,
  Equivalence,
  Linear_Order	:: "('a,'a)rel => bool"
  Chain		:: "[('a,'a)rel, ('a,'a)rel] => bool"	("(Chain'(_,_'))")

  (* Vectors *)
  Vector,
  Predicate,
  Coreflexive,
  Point		:: "('a,'b)rel => bool"


rules

  univalent_def	 "Univalent(R)	== R^R  I"
  total_def	 "Total(R)	== RL = L"
  function_def	 "Function(R)	== Univalent(R) & Total(R)"
  injective_def	 "Injective(R)	== Univalent(R^)"
  surjective_def "Surjective(R)	== Total(R^)"
  bijective_def	 "Bijective(R)	== Function(R^)"
  difunctional_def "Difunctional(R) == RR^R  R"

  reflexive_def	    "Reflexive(R)     == I  R"
  transitive_def    "Transitive(R)    == RR  R"
  preorder_def      "Preorder(R)      == Reflexive(R) & Transitive(R)"
  antisymmetric_def "Antisymmetric(R) == RR^  I"
  order_def         "Order(R)         == Preorder(R) & Antisymmetric(R)"
  symmetric_def     "Symmetric(R)     == R  R^"
  equivalence_def   "Equivalence(R)   == Preorder(R) & Symmetric(R)"
  linear_order_def  "Linear_Order(R)  == Order(R) & RR^ = L"
  chain_def         "Chain(R,Q)       == Total(R) & R^R  QQ^"

  vector_def	  "Vector(v)      == Lv = v"
  predicate_def	  "Predicate(b)   == bL = b"
  coreflexive_def "Coreflexive(e) == e  I"
  point_def	  "Point(p)       == Vector(p) & Function(p)"


end

